import random
import re
import time
from typing import Any, List, Optional, Tuple

import urwid

from zulipterminal.config.keys import commands_for_random_tips, is_command_key
from zulipterminal.config.symbols import (
    APPLICATION_TITLE_BAR_LINE,
    LIST_TITLE_BAR_LINE,
)
from zulipterminal.helper import WSL, asynch
from zulipterminal.ui_tools.boxes import SearchBox, WriteBox
from zulipterminal.ui_tools.views import (
    LeftColumnView,
    MiddleColumnView,
    RightColumnView,
)


class View(urwid.WidgetWrap):
    """
    A class responsible for providing the application's interface.
    """

    LEFT_WIDTH = 27
    RIGHT_WIDTH = 23

    def __init__(self, controller: Any) -> None:
        self.controller = controller
        self.palette = controller.theme
        self.model = controller.model
        self.users = self.model.users
        self.pinned_streams = self.model.pinned_streams
        self.unpinned_streams = self.model.unpinned_streams
        self.write_box = WriteBox(self)
        self.search_box = SearchBox(self.controller)

        self.message_view = None  # type: Any

        super().__init__(self.main_window())

    def left_column_view(self) -> Any:
        return LeftColumnView(View.LEFT_WIDTH, self)

    def middle_column_view(self) -> Any:
        self.middle_column = MiddleColumnView(self, self.model, self.write_box,
                                              self.search_box)
        return urwid.LineBox(self.middle_column, title='Messages',
                             tline=LIST_TITLE_BAR_LINE,
                             bline='', trcorner='│', tlcorner='│')

    def right_column_view(self) -> Any:
        self.users_view = RightColumnView(View.RIGHT_WIDTH, self)
        return urwid.LineBox(
            self.users_view, title="Users",
            tlcorner=LIST_TITLE_BAR_LINE,
            tline=LIST_TITLE_BAR_LINE,
            trcorner=LIST_TITLE_BAR_LINE,
            lline='', blcorner='─', rline='',
            bline='', brcorner=''
        )

    def get_random_help(self) -> List[Any]:
        # Get random allowed hotkey (ie. eligible for being displayed as a tip)
        allowed_commands = commands_for_random_tips()
        if not allowed_commands:
            return ['Help(?): ', ]
        random_command = random.choice(allowed_commands)
        return [
            'Help(?): ',
            ('code', ' ' + ', '.join(random_command['keys']) + ' '),
            ' ' + random_command['help_text'],
        ]

    @asynch
    def set_footer_text(self, text_list: Optional[List[Any]]=None,
                        duration: Optional[float]=None) -> None:
        if text_list is None:
            text = self.get_random_help()
        else:
            text = text_list
        self._w.footer.set_text(text)
        self.controller.update_screen()
        if duration is not None:
            assert duration > 0
            time.sleep(duration)
            self.set_footer_text()

    @asynch
    def set_typeahead_footer(self, suggestions: List[str],
                             state: Optional[int],
                             is_truncated: bool) -> None:
        if suggestions:
            # Wrap by space.
            footer_text = [' ' + s + ' '
                           for s in suggestions]  # type: List[Any]
            if state is not None:
                footer_text[state] = ('code', footer_text[state])
            if is_truncated:
                footer_text += [' [more] ']
            footer_text.insert(0, [' '])  # Add leading space.
        else:
            footer_text = [' [No matches found]']

        self.set_footer_text(footer_text)

    def footer_view(self) -> Any:
        text_header = self.get_random_help()
        return urwid.AttrWrap(urwid.Text(text_header), 'footer')

    def main_window(self) -> Any:
        self.left_panel = self.left_column_view()
        self.center_panel = self.middle_column_view()
        self.right_panel = self.right_column_view()
        if self.controller.autohide:
            body = [
                (View.LEFT_WIDTH, self.left_panel),
                ('weight', 10, self.center_panel),
                (0, self.right_panel),
            ]
        else:
            body = [
                (View.LEFT_WIDTH, self.left_panel),
                ('weight', 10, self.center_panel),
                (View.RIGHT_WIDTH, self.right_panel),
            ]
        self.body = urwid.Columns(body, focus_column=0)

        # NOTE: message_view is None, but middle_column_view is called above
        # and sets it.
        assert self.message_view is not None
        # NOTE: set_focus_changed_callback is actually called before the
        # focus is set, so the message is not read yet, it will be read when
        # the focus is changed again either vertically or horizontally.
        self.body._contents.set_focus_changed_callback(
            self.message_view.read_message)

        title_text = " {full_name} ({email}) - {server_name} ({url}) ".format(
                     full_name=self.model.user_full_name,
                     email=self.model.user_email,
                     server_name=self.model.server_name,
                     url=self.model.server_url)

        title_bar = urwid.Columns([
            urwid.Divider(div_char=APPLICATION_TITLE_BAR_LINE),
            (len(title_text), urwid.Text([title_text])),
            urwid.Divider(div_char=APPLICATION_TITLE_BAR_LINE),
        ])

        w = urwid.Frame(self.body, title_bar, focus_part='body',
                        footer=self.footer_view())
        return w

    def show_left_panel(self, *, visible: bool) -> None:
        if not self.controller.autohide:
            return
        width = View.LEFT_WIDTH if visible else 0
        self.body.contents[0] = (
            self.left_panel,
            self.body.options(width_type='given', width_amount=width),
        )
        if visible:
            self.body.focus_position = 0

    def show_right_panel(self, *, visible: bool) -> None:
        if not self.controller.autohide:
            return
        width = View.RIGHT_WIDTH if visible else 0
        self.body.contents[2] = (
            self.right_panel,
            self.body.options(width_type='given', width_amount=width),
        )
        if visible:
            self.body.focus_position = 2

    # FIXME: The type of size should be urwid_Size; this needs checking
    def keypress(self, size: Tuple[int, int], key: str) -> Optional[str]:
        self.model.new_user_input = True
        if self.controller.is_in_editor_mode():
            return self.controller.current_editor().keypress((size[1],), key)
        # Redirect commands to message_view.
        elif (is_command_key('SEARCH_MESSAGES', key)
                or is_command_key('NEXT_UNREAD_TOPIC', key)
                or is_command_key('NEXT_UNREAD_PM', key)
                or is_command_key('PRIVATE_MESSAGE', key)):
            self.body.focus_col = 1
            self.middle_column.keypress(size, key)
            return key
        elif is_command_key('ALL_PM', key):
            self.model.controller.show_all_pm(self)
            self.body.focus_col = 1
        elif is_command_key('ALL_STARRED', key):
            self.model.controller.show_all_starred(self)
            self.body.focus_col = 1
        elif is_command_key('ALL_MENTIONS', key):
            self.model.controller.show_all_mentions(self)
            self.body.focus_col = 1
        elif is_command_key('SEARCH_PEOPLE', key):
            # Start User Search if not in editor_mode
            self.body.focus_position = 2
            self.users_view.keypress(size, 'w')
            self.show_left_panel(visible=False)
            self.show_right_panel(visible=True)
            self.user_search.set_edit_text("")
            self.controller.enter_editor_mode_with(self.user_search)
            return key
        elif (is_command_key('SEARCH_STREAMS', key)
              or is_command_key('SEARCH_TOPICS', key)):
            # jump stream search
            self.body.focus_position = 0
            self.left_panel.keypress(size, 'q')
            self.show_right_panel(visible=False)
            self.show_left_panel(visible=True)
            if self.left_panel.is_in_topic_view:
                search_box = self.topic_w.topic_search_box
            else:
                search_box = self.stream_w.stream_search_box
            search_box.set_edit_text("")
            self.controller.enter_editor_mode_with(search_box)
            return key
        elif is_command_key('OPEN_DRAFT', key):
            saved_draft = self.model.session_draft_message()
            if saved_draft:
                if saved_draft['type'] == 'stream':
                    self.write_box.stream_box_view(
                        caption=saved_draft['display_recipient'],
                        title=saved_draft['subject'],
                        stream_id=saved_draft['stream_id'],
                    )
                elif saved_draft['type'] == 'private':
                    email_txt = saved_draft['display_recipient']
                    recipient_user_ids = [
                        self.model.user_dict[email.strip()]['user_id']
                        for email in email_txt.split(",")
                    ]
                    self.write_box.private_box_view(
                        email=email_txt,
                        recipient_user_ids=recipient_user_ids,
                    )
                content = saved_draft['content']
                self.write_box.msg_write_box.edit_text = content
                self.write_box.msg_write_box.edit_pos = len(content)
                self.middle_column.set_focus('footer')
            else:
                self.set_footer_text('No draft message was saved in'
                                     ' this session.', 3)
            return key
        elif is_command_key('ABOUT', key):
            self.controller.show_about()
            return key
        elif is_command_key('HELP', key):
            # Show help menu
            self.controller.show_help()
            return key
        # replace alternate keys with arrow/functional keys
        # This is needed for navigating in widgets
        # other than message_view.
        elif is_command_key('GO_UP', key):
            key = 'up'
        elif is_command_key('GO_DOWN', key):
            key = 'down'
        elif is_command_key('GO_LEFT', key):
            key = 'left'
        elif is_command_key('GO_RIGHT', key):
            key = 'right'
        elif is_command_key('SCROLL_UP', key):
            key = 'page up'
        elif is_command_key('SCROLL_DOWN', key):
            key = 'page down'
        elif is_command_key('GO_TO_BOTTOM', key):
            key = 'end'
        return super().keypress(size, key)


class Screen(urwid.raw_display.Screen):

    def write(self, data: Any) -> None:
        if WSL:
            # replace urwid's SI/SO, which produce artifacts under WSL.
            # https://github.com/urwid/urwid/issues/264#issuecomment-358633735
            # Above link describes the change.
            data = re.sub("[\x0e\x0f]", "", data)
        super().write(data)



import json
import time
from collections import OrderedDict, defaultdict
from concurrent.futures import Future, ThreadPoolExecutor, wait
from copy import deepcopy
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)
from urllib.parse import urlparse

import zulip
from typing_extensions import Literal, TypedDict

from zulipterminal import unicode_emojis
from zulipterminal.config.keys import primary_key_for_command
from zulipterminal.helper import (
    Message,
    NamedEmojiData,
    StreamData,
    asynch,
    canonicalize_color,
    classify_unread_counts,
    display_error_if_present,
    index_messages,
    initial_index,
    notify,
    set_count,
)
from zulipterminal.ui_tools.utils import create_msg_box_list


Event = TypedDict('Event', {
    'type': str,
    # typing:
    'sender': Dict[str, Any],  # 'email', ...
    # typing & reaction:
    'op': str,
    # reaction:
    'user': Dict[str, Any],  # 'email', 'user_id', 'full_name'
    'reaction_type': str,
    'emoji_code': str,
    'emoji_name': str,
    # reaction & update_message:
    'message_id': int,
    # update_message:
    'rendered_content': str,
    # update_message_flags:
    'messages': List[int],
    'operation': str,  # NOTE: deprecated in Zulip 4.0 / ZFL 32 -> 'op'
    'flag': str,
    'all': bool,
    # message:
    'message': Message,
    'flags': List[str],
    'subject': str,
    # subscription:
    'property': str,
    'user_id': int,  # Present when a streams subscribers are updated.
    'user_ids': List[int],  # NOTE: replaces 'user_id' in ZFL 35
    'stream_id': int,
    'stream_ids': List[int],  # NOTE: replaces 'stream_id' in ZFL 35 for peer*
    'value': bool,
    'message_ids': List[int]  # Present when subject of msg(s) is updated
}, total=False)  # Each Event will only have a subset of these

EditPropagateMode = Literal['change_one', 'change_all', 'change_later']

OFFLINE_THRESHOLD_SECS = 140


class ServerConnectionFailure(Exception):
    pass


def sort_streams(streams: List[StreamData]) -> None:
    """
    Used for sorting model.pinned_streams and model.unpinned_streams.
    """
    streams.sort(key=lambda s: s['name'].lower())


class Model:
    """
    A class responsible for storing the data to be displayed.
    """

    def __init__(self, controller: Any) -> None:
        self.controller = controller
        self.client = controller.client

        self.narrow = []  # type: List[Any]
        self._have_last_message = {}  # type: Dict[str, bool]
        self.stream_id = -1
        self.recipients = frozenset()  # type: FrozenSet[Any]
        self.index = initial_index

        self.user_id = -1
        self.user_email = ""
        self.user_full_name = ""
        self.server_url = '{uri.scheme}://{uri.netloc}/'.format(
                          uri=urlparse(self.client.base_url))
        self.server_name = ""

        self._notified_user_of_notification_failure = False

        self.event_actions = OrderedDict([
            ('message', self._handle_message_event),
            ('update_message', self._handle_update_message_event),
            ('reaction', self._handle_reaction_event),
            ('subscription', self._handle_subscription_event),
            ('typing', self._handle_typing_event),
            ('update_message_flags', self._handle_update_message_flags_event),
        ])  # type: OrderedDict[str, Callable[[Event], None]]

        self.initial_data = {}  # type: Dict[str, Any]

        # Register to the queue before initializing further so that we don't
        # lose any updates while messages are being fetched.
        self._update_initial_data()

        self.server_version = self.initial_data['zulip_version']
        self.server_feature_level = (
            self.initial_data.get('zulip_feature_level')
        )

        self.users = self.get_all_users()

        subscriptions = self.initial_data['subscriptions']
        stream_data = Model._stream_info_from_subscriptions(subscriptions)
        (self.stream_dict, self.muted_streams,
         self.pinned_streams, self.unpinned_streams) = stream_data

        # NOTE: The expected response has been upgraded from
        # [stream_name, topic] to [stream_name, topic, date_muted] in
        # feature level 1, server version 3.0.
        muted_topics = self.initial_data['muted_topics']
        assert set(map(len, muted_topics)) in (set(), {2}, {3})
        self._muted_topics = {
            (stream_name, topic): (None if self.server_feature_level is None
                                   else date_muted[0])
            for stream_name, topic, *date_muted in muted_topics
        }  # type: Dict[Tuple[str, str], Optional[int]]

        groups = self.initial_data['realm_user_groups']
        self.user_group_by_id = {}  # type: Dict[int, Dict[str, Any]]
        self.user_group_names = self._group_info_from_realm_user_groups(groups)

        self.unread_counts = classify_unread_counts(self)

        self._draft = None  # type: Optional[Message]
        unicode_emoji_data = unicode_emojis.EMOJI_DATA
        for name, data in unicode_emoji_data.items():
            data['type'] = 'unicode_emoji'
        typed_unicode_emoji_data = cast(NamedEmojiData, unicode_emoji_data)
        custom_emoji_data = self.fetch_custom_emojis()
        zulip_extra_emoji = {
                'zulip': {'code': 'zulip', 'type': 'zulip_extra_emoji'}
        }  # type: NamedEmojiData
        all_emoji_data = {**typed_unicode_emoji_data,
                          **custom_emoji_data,
                          **zulip_extra_emoji}.items()
        self.active_emoji_data = OrderedDict(sorted(all_emoji_data,
                                                    key=lambda e: e[0]))

        self.new_user_input = True
        self._start_presence_updates()

    def get_focus_in_current_narrow(self) -> Union[int, Set[None]]:
        """
        Returns the focus in the current narrow.
        For no existing focus this returns {}, otherwise the message ID.
        """
        return self.index['pointer'][repr(self.narrow)]

    def set_focus_in_current_narrow(self, focus_message: int) -> None:
        self.index['pointer'][repr(self.narrow)] = focus_message

    def is_search_narrow(self) -> bool:
        """
        Checks if the current narrow is a result of a previous search for
        a messages in a different narrow.
        """
        return 'search' in [subnarrow[0] for subnarrow in self.narrow]

    def set_narrow(self, *,
                   stream: Optional[str]=None,
                   topic: Optional[str]=None,
                   pms: bool=False,
                   pm_with: Optional[str]=None,
                   starred: bool=False,
                   mentioned: bool=False) -> bool:
        selected_params = {k for k, v in locals().items() if k != 'self' and v}
        valid_narrows = {
            frozenset(): [],
            frozenset(['stream']): [['stream', stream]],
            frozenset(['stream', 'topic']): [['stream', stream],
                                             ['topic', topic]],
            frozenset(['pms']): [['is', 'private']],
            frozenset(['pm_with']): [['pm_with', pm_with]],
            frozenset(['starred']): [['is', 'starred']],
            frozenset(['mentioned']): [['is', 'mentioned']],
        }  # type: Dict[FrozenSet[str], List[Any]]
        for narrow_param, narrow in valid_narrows.items():
            if narrow_param == selected_params:
                new_narrow = narrow
                break
        else:
            raise RuntimeError("Model.set_narrow parameters used incorrectly.")

        if new_narrow != self.narrow:
            self.narrow = new_narrow

            if pm_with is not None and new_narrow[0][0] == 'pm_with':
                users = pm_with.split(', ')
                self.recipients = frozenset(
                    [self.user_dict[user]['user_id'] for user in users]
                    + [self.user_id]
                )
            else:
                self.recipients = frozenset()
            return False
        else:
            return True

    def set_search_narrow(self, search_query: str) -> None:
        self.unset_search_narrow()
        self.narrow.append(['search', search_query])

    def unset_search_narrow(self) -> None:
        # If current narrow is a result of a previous started search,
        # we pop the ['search', 'text'] term in the narrow, before
        # setting a new narrow.
        if self.is_search_narrow():
            self.narrow = [item for item in self.narrow
                           if item[0] != 'search']

    def get_message_ids_in_current_narrow(self) -> Set[int]:
        narrow = self.narrow
        index = self.index
        if narrow == []:
            ids = index['all_msg_ids']
        elif self.is_search_narrow():  # Check searches first
            ids = index['search']
        elif narrow[0][0] == 'stream':
            stream_id = self.stream_id
            if len(narrow) == 1:
                ids = index['stream_msg_ids_by_stream_id'][stream_id]
            elif len(narrow) == 2:
                topic = narrow[1][1]
                ids = index['topic_msg_ids'][stream_id].get(topic, set())
        elif narrow[0][1] == 'private':
            ids = index['private_msg_ids']
        elif narrow[0][0] == 'pm_with':
            recipients = self.recipients
            ids = index['private_msg_ids_by_user_ids'].get(recipients, set())
        elif narrow[0][1] == 'starred':
            ids = index['starred_msg_ids']
        elif narrow[0][1] == 'mentioned':
            ids = index['mentioned_msg_ids']
        return ids.copy()

    def _notify_server_of_presence(self) -> Dict[str, Any]:
        response = self.client.update_presence(
                request={
                    # TODO: Determine `status` from terminal tab focus.
                    'status': 'active' if self.new_user_input else 'idle',
                    'new_user_input': self.new_user_input,
                }
            )
        self.new_user_input = False
        return response

    @asynch
    def _start_presence_updates(self) -> None:
        """
        Call `_notify_server_of_presence` every minute (version 1a).
        Use 'response' to update user list (version 1b).
        """
        # FIXME: Version 2: call endpoint with ping_only=True only when
        #        needed, and rely on presence events to update
        while True:
            response = self._notify_server_of_presence()
            if response['result'] == 'success':
                self.initial_data['presences'] = response['presences']
                self.users = self.get_all_users()
                if hasattr(self.controller, 'view'):
                    self.controller.view.users_view.update_user_list(
                        user_list=self.users)
            time.sleep(60)

    @asynch
    def react_to_message(self,
                         message: Message,
                         reaction_to_toggle: str) -> None:
        # FIXME Only support thumbs_up for now
        assert reaction_to_toggle == 'thumbs_up'

        reaction_to_toggle_spec = dict(
            emoji_name='thumbs_up',
            emoji_code='1f44d',
            reaction_type='unicode_emoji',
            message_id=str(message['id']))
        existing_reactions = [
            reaction['emoji_code']
            for reaction in message['reactions']
            if (reaction['user'].get('user_id', None) == self.user_id
                or reaction['user'].get('id', None) == self.user_id)
        ]
        if reaction_to_toggle_spec['emoji_code'] in existing_reactions:
            response = self.client.remove_reaction(reaction_to_toggle_spec)
        else:
            response = self.client.add_reaction(reaction_to_toggle_spec)
        display_error_if_present(response, self.controller)

    def session_draft_message(self) -> Optional[Message]:
        return deepcopy(self._draft)

    def save_draft(self, message: Message) -> None:
        self._draft = deepcopy(message)
        self.controller.view.set_footer_text("Saved message as draft", 3)

    @asynch
    def toggle_message_star_status(self, message: Message) -> None:
        base_request = dict(flag='starred', messages=[message['id']])
        if 'starred' in message['flags']:
            request = dict(base_request, op='remove')
        else:
            request = dict(base_request, op='add')
        response = self.client.update_message_flags(request)
        display_error_if_present(response, self.controller)

    @asynch
    def mark_message_ids_as_read(self, id_list: List[int]) -> None:
        if not id_list:
            return
        response = self.client.update_message_flags({
            'messages': id_list,
            'flag': 'read',
            'op': 'add',
        })
        display_error_if_present(response, self.controller)

    @asynch
    def send_typing_status_by_user_ids(self, recipient_user_ids: List[int],
                                       *, status: Literal['start', 'stop']
                                       ) -> None:
        if recipient_user_ids:
            request = {
                'to': recipient_user_ids,
                'op': status
            }
            response = self.client.set_typing_status(request)
            display_error_if_present(response, self.controller)
        else:
            raise RuntimeError('Empty recipient list.')

    def send_private_message(self, recipients: List[str],
                             content: str) -> bool:
        if recipients:
            request = {
                'type': 'private',
                'to': recipients,
                'content': content,
            }
            response = self.client.send_message(request)
            display_error_if_present(response, self.controller)
            return response['result'] == 'success'
        else:
            raise RuntimeError('Empty recipients list.')

    def send_stream_message(self, stream: str, topic: str,
                            content: str) -> bool:
        request = {
            'type': 'stream',
            'to': stream,
            'subject': topic,
            'content': content,
        }
        response = self.client.send_message(request)
        display_error_if_present(response, self.controller)
        return response['result'] == 'success'

    def update_private_message(self, msg_id: int, content: str) -> bool:
        request = {
            "message_id": msg_id,
            "content": content,
        }
        response = self.client.update_message(request)
        display_error_if_present(response, self.controller)
        return response['result'] == 'success'

    def update_stream_message(self, topic: str, message_id: int,
                              propagate_mode: EditPropagateMode,
                              content: Optional[str]=None) -> bool:
        request = {
            "message_id": message_id,
            "propagate_mode": propagate_mode,
            "topic": topic,
        }
        if content is not None:
            request['content'] = content

        response = self.client.update_message(request)
        display_error_if_present(response, self.controller)
        return response['result'] == 'success'

    def fetch_custom_emojis(self) -> NamedEmojiData:
        response = self.client.get_realm_emoji()
        custom_emojis = {emoji['name']: {'code': emoji_code,
                                         'type': 'realm_emoji'}
                         for emoji_code, emoji in response['emoji'].items()
                         if not emoji['deactivated']}  # type: NamedEmojiData
        display_error_if_present(response, self.controller)
        return custom_emojis

    def get_messages(self, *,
                     num_after: int, num_before: int,
                     anchor: Optional[int]) -> str:
        # anchor value may be specific message (int) or next unread (None)
        first_anchor = anchor is None
        anchor_value = anchor if anchor is not None else 0

        request = {
            'anchor': anchor_value,
            'num_before': num_before,
            'num_after': num_after,
            'apply_markdown': True,
            'use_first_unread_anchor': first_anchor,
            'client_gravatar': True,
            'narrow': json.dumps(self.narrow),
        }
        response = self.client.get_messages(message_filters=request)
        if response['result'] == 'success':
            self.index = index_messages(response['messages'], self, self.index)
            narrow_str = repr(self.narrow)
            if first_anchor and response['anchor'] != 10000000000000000:
                self.index['pointer'][narrow_str] = response['anchor']
            if 'found_newest' in response:
                self._have_last_message[narrow_str] = response['found_newest']
            else:
                # Older versions of the server does not contain the
                # 'found_newest' flag. Instead, we use this logic:
                query_range = num_after + num_before + 1
                self._have_last_message[narrow_str] = (
                    len(response['messages']) < query_range)
            return ""
        display_error_if_present(response, self.controller)
        return response['msg']

    def fetch_message_history(self, message_id: int,
                              ) -> List[Dict[str, Union[int, str]]]:
        """
        Fetches message edit history for a message using its ID.
        """
        response = self.client.get_message_history(message_id)
        if response['result'] == 'success':
            return response['message_history']
        display_error_if_present(response, self.controller)
        return list()

    def _fetch_topics_in_streams(self, stream_list: Iterable[int]) -> str:
        """
        Fetch all topics with specified stream_id's and
        index their names (Version 1)
        """
        # FIXME: Version 2: Fetch last 'n' recent topics for each stream.
        for stream_id in stream_list:
            response = self.client.get_stream_topics(stream_id)
            if response['result'] == 'success':
                self.index['topics'][stream_id] = [topic['name'] for
                                                   topic in response['topics']]
            else:
                display_error_if_present(response, self.controller)
                return response['msg']
        return ""

    def topics_in_stream(self, stream_id: int) -> List[str]:
        """
        Returns a list of topic names for stream_id from the index.
        """
        if not self.index['topics'][stream_id]:
            self._fetch_topics_in_streams([stream_id])

        return list(self.index['topics'][stream_id])

    @staticmethod
    def exception_safe_result(future: 'Future[str]') -> str:
        try:
            return future.result()
        except zulip.ZulipError as e:
            return str(e)

    def is_muted_stream(self, stream_id: int) -> bool:
        return stream_id in self.muted_streams

    def is_muted_topic(self, stream_id: int, topic: str) -> bool:
        """
        Returns True if topic is muted via muted_topics.
        """
        stream_name = self.stream_dict[stream_id]['name']
        topic_to_search = (stream_name, topic)
        return topic_to_search in self._muted_topics.keys()

    def _update_initial_data(self) -> None:
        # Thread Processes to reduce start time.
        # NOTE: Exceptions do not work well with threads
        with ThreadPoolExecutor(max_workers=1) as executor:
            futures = {
                'get_messages': executor.submit(self.get_messages,
                                                num_after=10,
                                                num_before=30,
                                                anchor=None),
                'register': executor.submit(self._register_desired_events,
                                            fetch_data=True),
            }  # type: Dict[str, Future[str]]

            # Wait for threads to complete
            wait(futures.values())

        results = {
            name: self.exception_safe_result(future)
            for name, future in futures.items()
        }  # type: Dict[str, str]
        if not any(results.values()):
            self.user_id = self.initial_data['user_id']
            self.user_email = self.initial_data['email']
            self.user_full_name = self.initial_data['full_name']
            self.server_name = self.initial_data['realm_name']
        else:
            failures = defaultdict(list)  # type: DefaultDict[str, List[str]]
            for name, result in results.items():
                if result:
                    failures[result].append(name)
            failure_text = [
                "{} ({})".format(error, ", ".join(sorted(calls)))
                for error, calls in failures.items()
            ]
            raise ServerConnectionFailure(", ".join(failure_text))

    def get_other_subscribers_in_stream(self, stream_id: Optional[int]=None,
                                        stream_name: Optional[str]=None,
                                        ) -> List[int]:
        assert stream_id is not None or stream_name is not None

        if stream_id:
            assert self.is_user_subscribed_to_stream(stream_id)

            return [sub
                    for sub in self.stream_dict[stream_id]['subscribers']
                    if sub != self.user_id]
        else:
            return [sub
                    for _, stream in self.stream_dict.items()
                    for sub in stream['subscribers']
                    if stream['name'] == stream_name
                    if sub != self.user_id]

    def get_all_users(self) -> List[Dict[str, Any]]:
        # Dict which stores the active/idle status of users (by email)
        presences = self.initial_data['presences']

        # Construct a dict of each user in the realm to look up by email
        # and a user-id to email mapping
        self.user_dict = dict()  # type: Dict[str, Dict[str, Any]]
        self.user_id_email_dict = dict()  # type: Dict[int, str]
        for user in self.initial_data['realm_users']:
            if self.user_id == user['user_id']:
                current_user = {
                    'full_name': user['full_name'],
                    'email': user['email'],
                    'user_id': user['user_id'],
                    'status': 'active',
                }
                continue
            email = user['email']
            if email in presences:  # presences currently subset of all users
                """
                * Aggregate our information on a user's presence across their
                * clients.
                *
                * For an explanation of the Zulip presence model this helps
                * implement, see the subsystem doc:
                https://zulip.readthedocs.io/en/latest/subsystems/presence.html
                *
                * This logic should match `status_from_timestamp` in the web
                * app's
                * `static/js/presence.js`.
                *
                * Out of the ClientPresence objects found in `presence`, we
                * consider only those with a timestamp newer than
                * OFFLINE_THRESHOLD_SECS; then of
                * those, return the one that has the greatest UserStatus, where
                * `active` > `idle` > `offline`.
                *
                * If there are several ClientPresence objects with the greatest
                * UserStatus, an arbitrary one is chosen.
                """
                aggregate_status = 'offline'
                for client in presences[email].items():
                    client_name = client[0]
                    status = client[1]['status']
                    timestamp = client[1]['timestamp']
                    if client_name == 'aggregated':
                        continue
                    elif (time.time() - timestamp) < OFFLINE_THRESHOLD_SECS:
                        if status == 'active':
                            aggregate_status = 'active'
                        if status == 'idle':
                            if aggregate_status != 'active':
                                aggregate_status = status
                        if status == 'offline':
                            if (aggregate_status != 'active'
                                    and aggregate_status != 'idle'):
                                aggregate_status = status

                status = aggregate_status
            else:
                # Set status of users not in the  `presence` list
                # as 'inactive'. They will not be displayed in the
                # user's list by default (only in the search list).
                status = 'inactive'
            self.user_dict[email] = {
                'full_name': user['full_name'],
                'email': email,
                'user_id': user['user_id'],
                'status': status,
            }
            self.user_id_email_dict[user['user_id']] = email

        # Add internal (cross-realm) bots to dicts
        for bot in self.initial_data['cross_realm_bots']:
            email = bot['email']
            self.user_dict[email] = {
                'full_name': bot['full_name'],
                'email': email,
                'user_id': bot['user_id'],
                'status': 'inactive',
            }
            self.user_id_email_dict[bot['user_id']] = email

        # Generate filtered lists for active & idle users
        active = [properties for properties in self.user_dict.values()
                  if properties['status'] == 'active']
        idle = [properties for properties in self.user_dict.values()
                if properties['status'] == 'idle']
        offline = [properties for properties in self.user_dict.values()
                   if properties['status'] == 'offline']
        inactive = [properties for properties in self.user_dict.values()
                    if properties['status'] == 'inactive']

        # Construct user_list from sorted components of each list
        user_list = sorted(active, key=lambda u: u['full_name'].casefold())
        user_list += sorted(idle, key=lambda u: u['full_name'].casefold())
        user_list += sorted(offline, key=lambda u: u['full_name'].casefold())
        user_list += sorted(inactive, key=lambda u: u['full_name'].casefold())
        # Add current user to the top of the list
        user_list.insert(0, current_user)
        self.user_dict[current_user['email']] = current_user
        self.user_id_email_dict[self.user_id] = current_user['email']

        return user_list

    def user_name_from_id(self, user_id: int) -> str:
        """
        Returns user's full name given their ID.
        """
        user_email = self.user_id_email_dict.get(user_id)

        if not user_email:
            raise RuntimeError('Invalid user ID.')

        return self.user_dict[user_email]['full_name']

    @staticmethod
    def _stream_info_from_subscriptions(
            subscriptions: List[Dict[str, Any]]
    ) -> Tuple[Dict[int, Any], Set[int], List[StreamData], List[StreamData]]:

        def make_reduced_stream_data(stream: Dict[str, Any]) -> StreamData:
            # stream_id has been changed to id.
            return StreamData({'name': stream['name'],
                               'id': stream['stream_id'],
                               'color': stream['color'],
                               'invite_only': stream['invite_only'],
                               'description': stream['description']})
        # Canonicalize color formats, since zulip server versions may use
        # different formats
        for subscription in subscriptions:
            subscription['color'] = canonicalize_color(subscription['color'])

        pinned_streams = [make_reduced_stream_data(stream)
                          for stream in subscriptions if stream['pin_to_top']]
        unpinned_streams = [make_reduced_stream_data(stream)
                            for stream in subscriptions
                            if not stream['pin_to_top']]
        sort_streams(pinned_streams)
        sort_streams(unpinned_streams)
        # Mapping of stream-id to all available stream info
        # Stream IDs for muted streams
        # Limited stream info sorted by name (used in display)
        return (
            {stream['stream_id']: stream for stream in subscriptions},
            {stream['stream_id'] for stream in subscriptions
             if stream['in_home_view'] is False},
            pinned_streams,
            unpinned_streams,
        )

    def _group_info_from_realm_user_groups(self,
                                           groups: List[Dict[str, Any]]
                                           ) -> List[str]:
        """
        Stores group information in the model and returns a list of
        group_names which helps in group typeahead. (Eg: @*terminal*)
        """
        for sub_group in groups:
            self.user_group_by_id[sub_group['id']] = {
                key: sub_group[key] for key in sub_group if key != 'id'}
        user_group_names = [self.user_group_by_id[group_id]['name']
                            for group_id in self.user_group_by_id]
        # Sort groups for typeahead to work alphabetically (case-insensitive)
        user_group_names.sort(key=str.lower)
        return user_group_names

    def toggle_stream_muted_status(self, stream_id: int) -> None:
        request = [{
            'stream_id': stream_id,
            'property': 'is_muted',
            'value': not self.is_muted_stream(stream_id)
            # True for muting and False for unmuting.
        }]
        response = self.client.update_subscription_settings(request)
        display_error_if_present(response, self.controller)

    def stream_id_from_name(self, stream_name: str) -> int:
        for stream_id, stream in self.stream_dict.items():
            if stream['name'] == stream_name:
                return stream_id
        raise RuntimeError("Invalid stream name.")

    def is_pinned_stream(self, stream_id: int) -> bool:
        return stream_id in [stream['id'] for stream in self.pinned_streams]

    def toggle_stream_pinned_status(self, stream_id: int) -> bool:
        request = [{
            'stream_id': stream_id,
            'property': 'pin_to_top',
            'value': not self.is_pinned_stream(stream_id)
        }]
        response = self.client.update_subscription_settings(request)
        return response['result'] == 'success'

    def is_user_subscribed_to_stream(self, stream_id: int) -> bool:
        return stream_id in self.stream_dict

    def _handle_subscription_event(self, event: Event) -> None:
        """
        Handle changes in subscription (eg. muting/unmuting,
                                        pinning/unpinning streams)
        """
        def get_stream_by_id(streams: List[StreamData], stream_id: int
                             ) -> StreamData:
            for stream in streams:
                if stream['id'] == stream_id:
                    return stream
            raise RuntimeError("Invalid stream id.")

        if event['op'] == 'update':
            if hasattr(self.controller, 'view'):
                if event.get('property', None) == 'in_home_view':
                    stream_id = event['stream_id']

                    # FIXME: Does this always contain the stream_id?
                    stream_button = (
                        self.controller.view.stream_id_to_button[stream_id]
                    )

                    unread_count = self.unread_counts['streams'][stream_id]
                    if event['value']:  # Unmuting streams
                        self.muted_streams.remove(stream_id)
                        self.unread_counts['all_msg'] += unread_count
                        stream_button.mark_unmuted(unread_count)
                    else:  # Muting streams
                        self.muted_streams.add(stream_id)
                        self.unread_counts['all_msg'] -= unread_count
                        stream_button.mark_muted()
                    self.controller.update_screen()
                elif event.get('property', None) == 'pin_to_top':
                    stream_id = event['stream_id']

                    # FIXME: Does this always contain the stream_id?
                    stream_button = (
                        self.controller.view.stream_id_to_button[stream_id]
                    )

                    if event['value']:
                        stream = get_stream_by_id(self.unpinned_streams,
                                                  stream_id)
                        if stream:
                            self.unpinned_streams.remove(stream)
                            self.pinned_streams.append(stream)
                    else:
                        stream = get_stream_by_id(self.pinned_streams,
                                                  stream_id)
                        if stream:
                            self.pinned_streams.remove(stream)
                            self.unpinned_streams.append(stream)
                    sort_streams(self.unpinned_streams)
                    sort_streams(self.pinned_streams)
                    self.controller.view.left_panel.update_stream_view()
                    self.controller.update_screen()
        elif event['op'] in ('peer_add', 'peer_remove'):
            # NOTE: ZFL 35 commit was not atomic with API change
            #       (ZFL >=35 can use new plural style)
            if 'stream_ids' not in event or 'user_ids' not in event:
                stream_ids = [event['stream_id']]
                user_ids = [event['user_id']]
            else:
                stream_ids = event['stream_ids']
                user_ids = event['user_ids']

            for stream_id in stream_ids:
                if self.is_user_subscribed_to_stream(stream_id):
                    subscribers = self.stream_dict[stream_id]['subscribers']
                    if event['op'] == 'peer_add':
                        subscribers.extend(user_ids)
                    else:
                        for user_id in user_ids:
                            subscribers.remove(user_id)

    def _handle_typing_event(self, event: Event) -> None:
        """
        Handle typing notifications (in private messages)
        """
        if hasattr(self.controller, 'view'):
            # If the user is in pm narrow with the person typing
            narrow = self.narrow
            if (len(narrow) == 1 and narrow[0][0] == 'pm_with'
                    and event['sender']['email'] in narrow[0][1].split(',')):
                if event['op'] == 'start':
                    user = self.user_dict[event['sender']['email']]
                    self.controller.view.set_footer_text([
                        ' ',
                        ('code', user['full_name']),
                        ' is typing...'
                    ])
                elif event['op'] == 'stop':
                    self.controller.view.set_footer_text()
                else:
                    raise RuntimeError("Unknown typing event operation")

    def get_invalid_recipient_emails(self, recipient_emails: List[str]
                                     ) -> List[str]:

        return [email for email in recipient_emails
                if email not in self.user_dict]

    def is_valid_stream(self, stream_name: str) -> bool:
        for stream in self.stream_dict.values():
            if stream['name'] == stream_name:
                return True
        return False

    def notify_user(self, message: Message) -> str:
        """
        return value signifies if notification failed, if it should occur
        """
        # Check if notifications are enabled by the user.
        # It is disabled by default.
        if not self.controller.notify_enabled:
            return ""
        if message['sender_id'] == self.user_id:
            return ""

        recipient = ''
        if message['type'] == 'private':
            target = 'you'
            if len(message['display_recipient']) > 2:
                extra_targets = [target] + [
                    recip['full_name']
                    for recip in message['display_recipient']
                    if recip['id'] not in (self.user_id, message['sender_id'])
                ]
                target = ', '.join(extra_targets)
            recipient = ' (to {})'.format(target)
        elif message['type'] == 'stream' and (
            {'mentioned', 'wildcard_mentioned'}.intersection(
                set(message['flags'])
            )
            or self.stream_dict[message['stream_id']]['desktop_notifications']
        ):
            recipient = ' (to {} -> {})'.format(message['display_recipient'],
                                                message['subject'])

        if recipient:
            return notify((self.server_name + ":\n"
                           + message['sender_full_name'] + recipient),
                          message['content'])
        return ""

    def _handle_message_event(self, event: Event) -> None:
        """
        Handle new messages (eg. add message to the end of the view)
        """
        message = event['message']
        # sometimes `flags` are missing in `event` so initialize
        # an empty list of flags in that case.
        message['flags'] = event.get('flags', [])
        # We need to update the topic order in index, unconditionally.
        if message['type'] == 'stream':
            # NOTE: The subsequent helper only updates the topic index based
            # on the message event not the UI (the UI is updated in a
            # consecutive block independently). However, it is critical to keep
            # the topics index synchronized as it used whenever the topics list
            # view is reconstructed later.
            self._update_topic_index(message['stream_id'],
                                     message['subject'])
            # If the topic view is toggled for incoming message's
            # recipient stream, then we re-arrange topic buttons
            # with most recent at the top.
            if hasattr(self.controller, 'view'):
                view = self.controller.view
                if (view.left_panel.is_in_topic_view_with_stream_id(
                        message['stream_id'])):
                    view.topic_w.update_topics_list(
                        message['stream_id'], message['subject'],
                        message['sender_id'])
                    self.controller.update_screen()

        # We can notify user regardless of whether UI is rendered or not,
        # but depend upon the UI to indicate failures.
        failed_command = self.notify_user(message)
        if (failed_command
                and hasattr(self.controller, 'view')
                and not self._notified_user_of_notification_failure):
            notice_template = (
                "You have enabled notifications, but your notification "
                "command '{}' could not be found."
                "\n\n"
                "The application will continue attempting to run this command "
                "in this session, but will not notify you again."
                "\n\n"
                "Press '{}' to close this window."
            )
            notice = notice_template.format(failed_command,
                                            primary_key_for_command("GO_BACK"))
            self.controller.popup_with_message(notice, width=50)
            self.controller.update_screen()
            self._notified_user_of_notification_failure = True

        # Index messages before calling set_count.
        self.index = index_messages([message], self, self.index)
        if 'read' not in message['flags']:
            set_count([message['id']], self.controller, 1)

        if (hasattr(self.controller, 'view')
                and self._have_last_message[repr(self.narrow)]):
            msg_log = self.controller.view.message_view.log
            if msg_log:
                last_message = msg_log[-1].original_widget.message
            else:
                last_message = None
            msg_w_list = create_msg_box_list(self, [message['id']],
                                             last_message=last_message)
            if not msg_w_list:
                return
            else:
                msg_w = msg_w_list[0]

            if not self.narrow:
                msg_log.append(msg_w)

            elif (self.narrow[0][1] == 'mentioned'
                    and 'mentioned' in message['flags']):
                msg_log.append(msg_w)

            elif (self.narrow[0][1] == message['type']
                    and len(self.narrow) == 1):
                msg_log.append(msg_w)

            elif (message['type'] == 'stream'
                    and self.narrow[0][0] == "stream"):
                recipient_stream = message['display_recipient']
                narrow_stream = self.narrow[0][1]
                append_to_stream = recipient_stream == narrow_stream

                if (append_to_stream
                    and (len(self.narrow) == 1
                         or (len(self.narrow) == 2
                             and self.narrow[1][1] == message['subject']))):
                    msg_log.append(msg_w)

            elif (message['type'] == 'private' and len(self.narrow) == 1
                    and self.narrow[0][0] == "pm_with"):
                narrow_recipients = self.recipients
                message_recipients = frozenset(
                    [user['id'] for user in message['display_recipient']])
                if narrow_recipients == message_recipients:
                    msg_log.append(msg_w)
            self.controller.update_screen()

    def _update_topic_index(self, stream_id: int, topic_name: str) -> None:
        """
        Update topic order in index based on incoming message.
        Helper method called by _handle_message_event
        """
        topic_list = self.topics_in_stream(stream_id)
        for topic_iterator, topic in enumerate(topic_list):
            if topic == topic_name:
                topic_list.insert(0, topic_list.pop(topic_iterator))
                break
        else:
            # No previous topics with same topic names are found
            # hence, it must be a new topic.
            topic_list.insert(0, topic_name)

        # Update the index.
        self.index['topics'][stream_id] = topic_list

    def _handle_update_message_event(self, event: Event) -> None:
        """
        Handle updated (edited) messages (changed content/subject)
        """
        # Update edited message status from single message id
        # NOTE: If all messages in topic have topic edited,
        #       they are not all marked as edited, as per server optimization
        message_id = event['message_id']
        indexed_message = self.index['messages'].get(message_id, None)

        if indexed_message:
            self.index['edited_messages'].add(message_id)

        # Update the rendered content, if the message is indexed
        if 'rendered_content' in event and indexed_message:
            indexed_message['content'] = event['rendered_content']
            self.index['messages'][message_id] = indexed_message
            self._update_rendered_view(message_id)

        # NOTE: This is independent of messages being indexed
        # Previous assertion:
        # * 'subject' is not present in update event if
        #   the event didn't have a 'subject' update.
        if 'subject' in event:
            new_subject = event['subject']
            stream_id = event['stream_id']

            # Update any indexed messages & re-render them
            for msg_id in event['message_ids']:
                indexed_msg = self.index['messages'].get(msg_id)
                if indexed_msg:
                    indexed_msg['subject'] = new_subject
                    self._update_rendered_view(msg_id)

            # If topic view is open, reload list else reset cache.
            if stream_id in self.index['topics']:
                if hasattr(self.controller, 'view'):
                    view = self.controller.view
                    if (view.left_panel.is_in_topic_view_with_stream_id(
                            stream_id)):
                        self._fetch_topics_in_streams([stream_id])
                        view.left_panel.show_topic_view(
                            view.topic_w.stream_button)
                        self.controller.update_screen()
                    else:
                        self.index['topics'][stream_id] = []

    def _handle_reaction_event(self, event: Event) -> None:
        """
        Handle change to reactions on a message
        """
        message_id = event['message_id']
        # If the message is indexed
        if self.index['messages'][message_id] != {}:

            message = self.index['messages'][message_id]
            if event['op'] == 'add':
                message['reactions'].append(
                    {
                        'user': event['user'],
                        'reaction_type': event['reaction_type'],
                        'emoji_code': event['emoji_code'],
                        'emoji_name': event['emoji_name'],
                    }
                )
            else:
                emoji_code = event['emoji_code']
                for reaction in message['reactions']:
                    # Since Who reacted is not displayed,
                    # remove the first one encountered
                    if reaction['emoji_code'] == emoji_code:
                        message['reactions'].remove(reaction)

            self.index['messages'][message_id] = message
            self._update_rendered_view(message_id)

    def _handle_update_message_flags_event(self, event: Event) -> None:
        """
        Handle change to message flags (eg. starred, read)
        """
        if (self.server_feature_level is None
                or self.server_feature_level < 32):
            operation = event['operation']
        else:
            operation = event['op']

        if event['all']:  # FIXME Should handle eventually
            return

        flag_to_change = event['flag']
        if flag_to_change not in {'starred', 'read'}:
            return

        if flag_to_change == 'read' and operation == 'remove':
            return

        indexed_message_ids = set(self.index['messages'])
        message_ids_to_mark = set(event['messages'])

        for message_id in message_ids_to_mark & indexed_message_ids:
            msg = self.index['messages'][message_id]
            if operation == 'add':
                if flag_to_change not in msg['flags']:
                    msg['flags'].append(flag_to_change)
            elif operation == 'remove':
                if flag_to_change in msg['flags']:
                    msg['flags'].remove(flag_to_change)
            else:
                raise RuntimeError(event, msg['flags'])

            self.index['messages'][message_id] = msg
            self._update_rendered_view(message_id)

        if operation == 'add' and flag_to_change == 'read':
            set_count(list(message_ids_to_mark & indexed_message_ids),
                      self.controller, -1)

    def _update_rendered_view(self, msg_id: int) -> None:
        """
        Helper method called by various _handle_* methods
        """
        # Update new content in the rendered view
        view = self.controller.view
        for msg_w in view.message_view.log:
            msg_box = msg_w.original_widget
            if msg_box.message['id'] == msg_id:
                # Remove the message if it no longer belongs in the current
                # narrow.
                if (len(self.narrow) == 2
                        and msg_box.message['subject'] != self.narrow[1][1]):
                    view.message_view.log.remove(msg_w)
                    # Change narrow if there are no messages left in the
                    # current narrow.
                    if not view.message_view.log:
                        msg_w_list = create_msg_box_list(
                                        self, [msg_id],
                                        last_message=msg_box.last_message)
                        if msg_w_list:
                            self.controller.narrow_to_topic(
                                msg_w_list[0].original_widget)
                    self.controller.update_screen()
                    return

                msg_w_list = create_msg_box_list(
                                self, [msg_id],
                                last_message=msg_box.last_message)
                if not msg_w_list:
                    return
                else:
                    new_msg_w = msg_w_list[0]
                    msg_pos = view.message_view.log.index(msg_w)
                    view.message_view.log[msg_pos] = new_msg_w

                    # If this is not the last message in the view
                    # update the next message's last_message too.
                    if len(view.message_view.log) != (msg_pos + 1):
                        next_msg_w = view.message_view.log[msg_pos + 1]
                        msg_w_list = create_msg_box_list(
                            self, [next_msg_w.original_widget.message['id']],
                            last_message=new_msg_w.original_widget.message)
                        view.message_view.log[msg_pos + 1] = msg_w_list[0]
                    self.controller.update_screen()
                    return

    def _register_desired_events(self, *, fetch_data: bool=False) -> str:
        fetch_types = None if not fetch_data else [
            'realm',
            'presence',
            'subscription',
            'message',
            'update_message_flags',
            'muted_topics',
            'realm_user',  # Enables cross_realm_bots
            'realm_user_groups',
            # zulip_version and zulip_feature_level are always returned in
            # POST /register from Feature level 3.
            'zulip_version',
        ]
        event_types = list(self.event_actions)
        try:
            response = self.client.register(event_types=event_types,
                                            fetch_event_types=fetch_types,
                                            client_gravatar=True,
                                            apply_markdown=True,
                                            include_subscribers=True)
        except zulip.ZulipError as e:
            return str(e)

        if response['result'] == 'success':
            if fetch_data:
                self.initial_data.update(response)
            self.max_message_id = response['max_message_id']
            self.queue_id = response['queue_id']
            self.last_event_id = response['last_event_id']
            return ""
        return response['msg']

    @asynch
    def poll_for_events(self) -> None:
        reregister_timeout = 10
        queue_id = self.queue_id
        last_event_id = self.last_event_id
        while True:
            if queue_id is None:
                while True:
                    if not self._register_desired_events():
                        queue_id = self.queue_id
                        last_event_id = self.last_event_id
                        break
                    time.sleep(reregister_timeout)

            response = self.client.get_events(
                queue_id=queue_id,
                last_event_id=last_event_id
            )

            if 'error' in response['result']:
                if response["msg"].startswith("Bad event queue id:"):
                    # Our event queue went away, probably because
                    # we were asleep or the server restarted
                    # abnormally.  We may have missed some
                    # events while the network was down or
                    # something, but there's not really anything
                    # we can do about it other than resuming
                    # getting new ones.
                    #
                    # Reset queue_id to register a new event queue.
                    queue_id = None
                time.sleep(1)
                continue

            for event in response['events']:
                last_event_id = max(last_event_id, int(event['id']))
                if event['type'] in self.event_actions:
                    try:
                        self.event_actions[event['type']](event)
                    except Exception:
                        import sys
                        (self.controller.
                         raise_exception_in_main_thread(sys.exc_info(),
                                                        critical=False))



from typing import Any, Iterable, List, Optional, Union

import urwid

from zulipterminal.helper import Message
from zulipterminal.ui_tools.boxes import MessageBox


def create_msg_box_list(model: Any, messages: Union[None, Iterable[Any]]=None,
                        *,
                        focus_msg_id: Union[None, int]=None,
                        last_message: Optional[Message]=None) -> List[Any]:
    """
    MessageBox for every message displayed is created here.
    """
    if not model.narrow and messages is None:
        messages = list(model.index['all_msg_ids'])
    if messages is not None:
        message_list = [model.index['messages'][id] for id in messages]
    message_list.sort(key=lambda msg: msg['timestamp'])
    w_list = []
    focus_msg = None
    last_msg = last_message
    muted_msgs = 0  # No of messages that are muted.
    for msg in message_list:
        if is_unsubscribed_message(msg, model):
            continue
        # Remove messages of muted topics / streams.
        if is_muted(msg, model):
            muted_msgs += 1
            if model.narrow == []:  # Don't show in 'All messages'.
                continue
        msg_flag = 'unread'  # type: Union[str, None]
        flags = msg.get('flags')
        # update_messages sends messages with no flags
        # but flags are set to [] when fetching old messages.
        if flags and ('read' in flags):
            msg_flag = None
        elif focus_msg is None:
            focus_msg = message_list.index(msg) - muted_msgs
        if msg['id'] == focus_msg_id:
            focus_msg = message_list.index(msg) - muted_msgs
        w_list.append(urwid.AttrMap(
                    MessageBox(msg, model, last_msg),
                    msg_flag,
                    'msg_selected'
        ))
        last_msg = msg
    if focus_msg is not None:
        model.set_focus_in_current_narrow(focus_msg)
    return w_list


def is_muted(msg: Message, model: Any) -> bool:
    # PMs cannot be muted
    if msg['type'] == 'private':
        return False
    # In a topic narrow
    elif len(model.narrow) == 2:
        return False
    elif model.is_muted_stream(msg['stream_id']):
        return True
    elif model.is_muted_topic(msg['stream_id'], msg['subject']):
        return True
    return False


def is_unsubscribed_message(msg: Message, model: Any) -> bool:
    if msg['type'] == 'private':
        return False
    if not model.is_user_subscribed_to_stream(msg['stream_id']):
        return True
    return False



import os
import signal
import sys
import time
from collections import OrderedDict
from functools import partial
from platform import platform
from types import TracebackType
from typing import Any, List, Optional, Tuple, Type

import urwid
import zulip
from typing_extensions import Literal

from zulipterminal.config.themes import ThemeSpec
from zulipterminal.helper import Message, asynch
from zulipterminal.model import Model
from zulipterminal.ui import Screen, View
from zulipterminal.ui_tools.utils import create_msg_box_list
from zulipterminal.ui_tools.views import (
    AboutView,
    EditHistoryView,
    EditModeView,
    HelpView,
    MsgInfoView,
    NoticeView,
    PopUpConfirmationView,
    StreamInfoView,
    StreamMembersView,
)
from zulipterminal.version import ZT_VERSION


ExceptionInfo = Tuple[Type[BaseException], BaseException, TracebackType]


class Controller:
    """
    A class responsible for setting up the model and view and running
    the application.
    """

    def __init__(self, config_file: str, theme_name: str, theme: ThemeSpec,
                 color_depth: int, in_explore_mode: bool,
                 autohide: bool, notify: bool, footlinks: bool) -> None:
        self.theme_name = theme_name
        self.theme = theme
        self.color_depth = color_depth
        self.in_explore_mode = in_explore_mode
        self.autohide = autohide
        self.notify_enabled = notify
        self.footlinks_enabled = footlinks

        self._editor = None  # type: Optional[Any]

        self.show_loading()
        self.client = zulip.Client(config_file=config_file,
                                   client='ZulipTerminal/{} {}'.
                                          format(ZT_VERSION, platform()))
        self.model = Model(self)
        self.view = View(self)
        # Start polling for events after view is rendered.
        self.model.poll_for_events()

        screen = Screen()
        screen.set_terminal_properties(colors=self.color_depth)
        self.loop = urwid.MainLoop(self.view,
                                   self.theme,
                                   screen=screen)

        # urwid pipe for concurrent screen update handling
        self._update_pipe = self.loop.watch_pipe(self._draw_screen)

        # data and urwid pipe for inter-thread exception handling
        self._exception_info = None  # type: Optional[ExceptionInfo]
        self._critical_exception = False
        self._exception_pipe = self.loop.watch_pipe(self._raise_exception)

        # Register new ^C handler
        signal.signal(signal.SIGINT, self.exit_handler)

    def raise_exception_in_main_thread(self,
                                       exc_info: ExceptionInfo,
                                       *, critical: bool) -> None:
        """
        Sets an exception from another thread, which is cleanly handled
        from within the Controller thread via _raise_exception
        """
        # Exceptions shouldn't occur before the pipe is set
        assert hasattr(self, '_exception_pipe')

        if isinstance(exc_info, tuple):
            self._exception_info = exc_info
            self._critical_exception = critical
        else:
            self._exception_info = (
                RuntimeError,
                "Invalid cross-thread exception info '{}'".format(exc_info),
                None
            )
            self._critical_exception = True
        os.write(self._exception_pipe, b'1')

    def is_in_editor_mode(self) -> bool:
        return self._editor is not None

    def enter_editor_mode_with(self, editor: Any) -> None:
        assert self._editor is None, "Already in editor mode"
        self._editor = editor

    def exit_editor_mode(self) -> None:
        self._editor = None

    def current_editor(self) -> Any:
        assert self._editor is not None, "Current editor is None"
        return self._editor

    @asynch
    def show_loading(self) -> None:

        def spinning_cursor() -> Any:
            while True:
                yield from '|/-\\'

        spinner = spinning_cursor()
        sys.stdout.write("\033[92mWelcome to Zulip.\033[0m\n")
        while not hasattr(self, 'view'):
            next_spinner = "Loading " + next(spinner)
            sys.stdout.write(next_spinner)
            sys.stdout.flush()
            time.sleep(0.1)
            sys.stdout.write('\b' * len(next_spinner))

        self.capture_stdout()

    def capture_stdout(self, path: str='debug.log') -> None:
        if hasattr(self, '_stdout'):
            return

        self._stdout = sys.stdout
        sys.stdout = open(path, 'a')

    def restore_stdout(self) -> None:
        if not hasattr(self, '_stdout'):
            return

        sys.stdout.flush()
        sys.stdout.close()
        sys.stdout = self._stdout
        sys.stdout.write('\n')
        del self._stdout

    def update_screen(self) -> None:
        # Update should not happen until pipe is set
        assert hasattr(self, '_update_pipe')
        # Write something to update pipe to trigger draw_screen
        os.write(self._update_pipe, b'1')

    def _draw_screen(self, *args: Any, **kwargs: Any) -> Literal[True]:
        self.loop.draw_screen()
        return True  # Always retain pipe

    def maximum_popup_dimensions(self) -> Tuple[int, int]:
        """
        Returns 3/4th of the screen estate's columns and rows.
        """
        max_cols, max_rows = map(lambda num: 3 * num // 4,
                                 self.loop.screen.get_cols_rows())
        return max_cols, max_rows

    def show_pop_up(self, to_show: Any, style: str) -> None:
        border_lines = dict(tlcorner='▛', tline='▀', trcorner='▜',
                            rline='▐', lline='▌',
                            blcorner='▙', bline='▄', brcorner='▟')
        text = urwid.Text(to_show.title, align='center')
        title_map = urwid.AttrMap(urwid.Filler(text), style)
        title_box_adapter = urwid.BoxAdapter(title_map, height=1)
        title_box = urwid.LineBox(
            title_box_adapter, tlcorner='▄', tline='▄', trcorner='▄',
            rline='', lline='', blcorner='', bline='', brcorner=''
        )
        title = urwid.AttrMap(title_box, 'popup_border')
        content = urwid.LineBox(to_show, **border_lines)
        self.loop.widget = urwid.Overlay(
            urwid.AttrMap(urwid.Frame(header=title, body=content),
                          'popup_border'),
            self.view,
            align='center',
            valign='middle',
            # +2 to both of the following, due to LineBox
            # +2 to height, due to title enhancement
            width=to_show.width + 2,
            height=to_show.height + 4,
        )

    def exit_popup(self) -> None:
        self.loop.widget = self.view

    def show_help(self) -> None:
        help_view = HelpView(self, "Help Menu (up/down scrolls)")
        self.show_pop_up(help_view, 'area:help')

    def show_topic_edit_mode(self, button: Any) -> None:
        self.show_pop_up(EditModeView(self, button), 'area:msg')

    def show_msg_info(self, msg: Message,
                      message_links: 'OrderedDict[str, Tuple[str, int, bool]]',
                      time_mentions: List[Tuple[str, str]],
                      ) -> None:
        msg_info_view = MsgInfoView(self, msg,
                                    "Message Information (up/down scrolls)",
                                    message_links, time_mentions)
        self.show_pop_up(msg_info_view, 'area:msg')

    def show_stream_info(self, stream_id: int) -> None:
        show_stream_view = StreamInfoView(self, stream_id)
        self.show_pop_up(show_stream_view, 'area:stream')

    def show_stream_members(self, stream_id: int) -> None:
        stream_members_view = StreamMembersView(self, stream_id)
        self.show_pop_up(stream_members_view, 'area:stream')

    def popup_with_message(self, text: str, width: int) -> None:
        self.show_pop_up(NoticeView(self, text, width, "NOTICE"),
                         'area:error')

    def show_about(self) -> None:
        self.show_pop_up(
            AboutView(self, 'About', zt_version=ZT_VERSION,
                      server_version=self.model.server_version,
                      server_feature_level=self.model.server_feature_level,
                      theme_name=self.theme_name, color_depth=self.color_depth,
                      autohide_enabled=self.autohide,
                      footlink_enabled=self.footlinks_enabled),
            'area:help'
        )

    def show_edit_history(
        self, message: Message,
        message_links: 'OrderedDict[str, Tuple[str, int, bool]]',
        time_mentions: List[Tuple[str, str]],
    ) -> None:
        self.show_pop_up(
            EditHistoryView(self, message, message_links, time_mentions,
                            'Edit History (up/down scrolls)'),
            'area:msg'
        )

    def search_messages(self, text: str) -> None:
        # Search for a text in messages
        self.model.index['search'].clear()
        self.model.set_search_narrow(text)

        self.model.get_messages(num_after=0, num_before=30, anchor=10000000000)
        msg_id_list = self.model.get_message_ids_in_current_narrow()

        w_list = create_msg_box_list(self.model, msg_id_list)
        self.view.message_view.log.clear()
        self.view.message_view.log.extend(w_list)
        focus_position = 0
        if 0 <= focus_position < len(w_list):
            self.view.message_view.set_focus(focus_position)

    def save_draft_confirmation_popup(self, message: Message) -> None:
        question = urwid.Text('Save this message as a draft?'
                              ' (This will overwrite the existing draft.)')
        save_draft = partial(self.model.save_draft, message)
        self.loop.widget = PopUpConfirmationView(self, question, save_draft)

    def stream_muting_confirmation_popup(self, button: Any) -> None:
        currently_muted = self.model.is_muted_stream(button.stream_id)
        type_of_action = "unmuting" if currently_muted else "muting"
        question = urwid.Text(("bold", "Confirm " + type_of_action
                               + " of stream '" + button.stream_name + "' ?"),
                              "center")
        mute_this_stream = partial(self.model.toggle_stream_muted_status,
                                   button.stream_id)
        self.loop.widget = PopUpConfirmationView(self, question,
                                                 mute_this_stream)

    def _narrow_to(self, button: Any, anchor: Optional[int],
                   **narrow: Any) -> None:
        already_narrowed = self.model.set_narrow(**narrow)
        if already_narrowed:
            return

        # store the steam id in the model (required for get_message_ids...)
        if hasattr(button, 'stream_id'):  # FIXME Include in set_narrow?
            self.model.stream_id = button.stream_id

        msg_id_list = self.model.get_message_ids_in_current_narrow()

        # if no messages are found get more messages
        if len(msg_id_list) == 0:
            self.model.get_messages(num_before=30,
                                    num_after=10,
                                    anchor=anchor)
            msg_id_list = self.model.get_message_ids_in_current_narrow()

        w_list = create_msg_box_list(self.model,
                                     msg_id_list,
                                     focus_msg_id=anchor)

        focus_position = self.model.get_focus_in_current_narrow()
        if focus_position == set():  # No available focus; set to end
            focus_position = len(w_list) - 1
        assert not isinstance(focus_position, set)

        self.view.message_view.log.clear()
        if 0 <= focus_position < len(w_list):
            self.view.message_view.log.extend(w_list, focus_position)
        else:
            self.view.message_view.log.extend(w_list)

        self.exit_editor_mode()

    def narrow_to_stream(self, button: Any) -> None:
        if hasattr(button, 'message'):
            anchor = button.message['id']
        else:
            anchor = None

        self._narrow_to(button,
                        anchor=anchor,
                        stream=button.stream_name)

    def narrow_to_topic(self, button: Any) -> None:
        if hasattr(button, 'message'):
            anchor = button.message['id']
        else:
            anchor = None

        self._narrow_to(button,
                        anchor=anchor,
                        stream=button.stream_name,
                        topic=button.topic_name)

    def narrow_to_user(self, button: Any) -> None:
        if hasattr(button, 'message'):
            user_emails = button.recipients_emails
            anchor = button.message['id']
        else:
            user_emails = button.email
            anchor = None

        self._narrow_to(button,
                        anchor=anchor,
                        pm_with=user_emails)

    def show_all_messages(self, button: Any) -> None:
        if hasattr(button, 'message'):
            anchor = button.message['id']
        else:
            anchor = None

        self._narrow_to(button,
                        anchor=anchor)

    def show_all_pm(self, button: Any) -> None:
        if hasattr(button, 'message'):
            anchor = button.message['id']
        else:
            anchor = None

        self._narrow_to(button,
                        anchor=anchor,
                        pms=True)

    def show_all_starred(self, button: Any) -> None:
        # NOTE: Should we ensure we maintain anchor focus here?
        # (it seems to work fine without)
        self._narrow_to(button,
                        anchor=None,
                        starred=True)

    def show_all_mentions(self, button: Any) -> None:
        # NOTE: Should we ensure we maintain anchor focus here?
        # (As with starred, it seems to work fine without)
        self._narrow_to(button,
                        anchor=None,
                        mentioned=True)

    def deregister_client(self) -> None:
        queue_id = self.model.queue_id
        self.client.deregister(queue_id, 1.0)

    def exit_handler(self, signum: int, frame: Any) -> None:
        self.deregister_client()
        sys.exit(0)

    def _raise_exception(self, *args: Any, **kwargs: Any) -> Literal[True]:
        if self._exception_info is not None:
            exc = self._exception_info
            if self._critical_exception:
                raise exc[0].with_traceback(exc[1], exc[2])
            else:
                import traceback
                exception_logfile = "zulip-terminal-thread-exceptions.log"
                with open(exception_logfile, "a") as logfile:
                    traceback.print_exception(*exc, file=logfile)
                message = (
                    "An exception occurred:"
                    + "\n\n"
                    + "".join(traceback.format_exception_only(exc[0], exc[1]))
                    + "\n"
                    + "The application should continue functioning, but you "
                    + "may notice inconsistent behavior in this session."
                    + "\n\n"
                    + "Please report this to us either in"
                    + "\n"
                    + "* the #zulip-terminal stream"
                    + "\n"
                    + "  (https://chat.zulip.org/#narrow/stream/"
                    + "206-zulip-terminal in the webapp)"
                    + "\n"
                    + "* an issue at "
                    + "https://github.com/zulip/zulip-terminal/issues"
                    + "\n\n"
                    + "Details of the exception can be found in "
                    + exception_logfile
                )
                self.popup_with_message(message, width=80)
                self._exception_info = None
        return True  # If don't raise, retain pipe

    def main(self) -> None:
        try:
            # TODO: Enable resuming? (in which case, remove ^Z below)
            disabled_keys = {
                'susp': 'undefined',  # Disable ^Z - no suspending
                'stop': 'undefined',  # Disable ^S - enabling shortcut key use
                'quit': 'undefined',  # Disable ^\, ^4
            }
            old_signal_list = self.loop.screen.tty_signal_keys(**disabled_keys)
            self.loop.run()

        except Exception:
            self.restore_stdout()
            self.loop.screen.tty_signal_keys(*old_signal_list)
            raise

        finally:
            self.restore_stdout()
            self.loop.screen.tty_signal_keys(*old_signal_list)



import re
from typing import Any, Callable, Dict, Optional, Tuple, Union, cast
from urllib.parse import urljoin, urlparse

import urwid
from typing_extensions import TypedDict

from zulipterminal.config.keys import is_command_key, primary_key_for_command
from zulipterminal.config.symbols import (
    MUTE_MARKER,
    STREAM_MARKER_PRIVATE,
    STREAM_MARKER_PUBLIC,
)
from zulipterminal.helper import (
    StreamData,
    edit_mode_captions,
    hash_util_decode,
)
from zulipterminal.urwid_types import urwid_Size


class MenuButton(urwid.Button):
    def __init__(self, caption: Any, email: str='') -> None:
        self.caption = caption  # str
        self.email = email
        super().__init__("")
        self._w = urwid.AttrMap(urwid.SelectableIcon(
            [self.caption], 0), None, 'selected')


class TopButton(urwid.Button):
    def __init__(self, controller: Any, caption: str,
                 show_function: Callable[..., Any], width: int,
                 prefix_character: Union[str, Tuple[Any, str]]='\N{BULLET}',
                 text_color: Optional[str]=None,
                 count: int=0) -> None:
        if isinstance(prefix_character, tuple):
            prefix = prefix_character[1]
        else:
            prefix = prefix_character
        assert len(prefix) in (0, 1)
        self._caption = caption
        self.prefix_character = prefix_character
        self.post_prefix_spacing = ' ' if prefix else ''
        self.count = count

        prefix_length = 0 if prefix == '' else 2
        # Space either side, at least one space between
        self.width_for_text_and_count = width - 3 - prefix_length

        self.original_color = text_color
        self.show_function = show_function
        super().__init__("")
        self.update_count(count, text_color)
        self.controller = controller
        urwid.connect_signal(self, 'click', self.activate)

    def update_count(self, count: int, text_color: Optional[str]=None) -> None:
        new_color = self.original_color if text_color is None else text_color

        self.count = count
        if count == 0:
            count_text = ''
        else:
            count_text = str(count)
        self.update_widget(('unread_count', count_text), new_color)

    def update_widget(self, count_text: Tuple[str, str],
                      text_color: Optional[str]) -> Any:
        # Note that we don't modify self._caption
        max_caption_length = (self.width_for_text_and_count
                              - len(count_text[1]))
        if len(self._caption) > max_caption_length:
            caption = (self._caption[:max_caption_length - 1]
                       + '\N{HORIZONTAL ELLIPSIS}')
        else:
            caption = self._caption
        num_extra_spaces = (
            self.width_for_text_and_count - len(count_text[1]) - len(caption)
        )

        # NOTE: Generated text does not include space at end
        self._w = urwid.AttrMap(urwid.SelectableIcon(
            [' ', self.prefix_character, self.post_prefix_spacing,
             '{}{}'.format(caption, num_extra_spaces * ' '),
             ' ', count_text],
            self.width_for_text_and_count + 5),  # cursor location
            text_color,
            'selected')

    def activate(self, key: Any) -> None:
        self.controller.view.show_left_panel(visible=False)
        self.controller.view.show_right_panel(visible=False)
        self.controller.view.body.focus_col = 1
        self.show_function(self)

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('ENTER', key):
            self.activate(key)
            return None
        else:  # This is in the else clause, to avoid multiple activation
            return super().keypress(size, key)


class HomeButton(TopButton):
    def __init__(self, controller: Any, width: int, count: int=0) -> None:
        button_text = ("All messages     ["
                       + primary_key_for_command("ALL_MESSAGES")
                       + "]")
        super().__init__(controller, button_text,
                         controller.show_all_messages, count=count,
                         prefix_character='',
                         width=width)


class PMButton(TopButton):
    def __init__(self, controller: Any, width: int, count: int=0) -> None:
        button_text = ("Private messages ["
                       + primary_key_for_command("ALL_PM")
                       + "]")
        super().__init__(controller, button_text,
                         controller.show_all_pm, count=count,
                         prefix_character='',
                         width=width)


class MentionedButton(TopButton):
    def __init__(self, controller: Any, width: int, count: int=0) -> None:
        button_text = ("Mentions         ["
                       + primary_key_for_command("ALL_MENTIONS")
                       + "]")
        super().__init__(controller, button_text,
                         controller.show_all_mentions,
                         width=width,
                         count=count,
                         prefix_character='')


class StarredButton(TopButton):
    def __init__(self, controller: Any, width: int) -> None:
        button_text = ("Starred messages ["
                       + primary_key_for_command("ALL_STARRED")
                       + "]")
        super().__init__(controller, button_text,
                         controller.show_all_starred,
                         width=width,
                         prefix_character='',
                         count=0)  # Starred messages are already marked read


class StreamButton(TopButton):
    def __init__(self, properties: StreamData,
                 controller: Any, view: Any, width: int,
                 count: int=0) -> None:
        # FIXME Is having self.stream_id the best way to do this?
        # (self.stream_id is used elsewhere)
        self.stream_name = properties['name']
        self.stream_id = properties['id']
        self.color = properties['color']
        is_private = properties['invite_only']
        self.description = properties['description']

        self.model = controller.model
        self.count = count
        self.view = view

        for entry in view.palette:
            if entry[0] is None:
                background = entry[5] if len(entry) > 4 else entry[2]
                inverse_text = background if background else 'black'
                break
        view.palette.append((
            self.color, '', '', 'bold', self.color + ', bold', background))
        view.palette.append((
            's' + self.color, '', '', 'standout', inverse_text, self.color))

        stream_marker = (STREAM_MARKER_PRIVATE if is_private
                         else STREAM_MARKER_PUBLIC)
        super().__init__(controller,
                         caption=self.stream_name,
                         show_function=controller.narrow_to_stream,
                         prefix_character=(self.color, stream_marker),
                         width=width,
                         count=count)

        # Mark muted streams 'M' during button creation.
        if self.model.is_muted_stream(self.stream_id):
            self.mark_muted()

    def mark_muted(self) -> None:
        self.update_widget(('muted', MUTE_MARKER), 'muted')
        self.view.home_button.update_count(
            self.model.unread_counts['all_msg'])

    def mark_unmuted(self, unread_count: int) -> None:
        self.update_count(unread_count)
        self.view.home_button.update_count(
            self.model.unread_counts['all_msg'])

    def keypress(self, size: urwid_Size, key: str) -> Optional[str]:
        if is_command_key('TOGGLE_TOPIC', key):
            self.view.left_panel.show_topic_view(self)
        elif is_command_key('TOGGLE_MUTE_STREAM', key):
            self.controller.stream_muting_confirmation_popup(self)
        elif is_command_key('STREAM_DESC', key):
            self.model.controller.show_stream_info(self.stream_id)
        return super().keypress(size, key)


class UserButton(TopButton):
    def __init__(self, user: Dict[str, Any], controller: Any,
                 view: Any, width: int,
                 state_marker: str,
                 color: Optional[str]=None, count: int=0,
                 is_current_user: bool=False) -> None:
        # Properties accessed externally
        self.email = user['email']
        self.user_id = user['user_id']

        self._view = view  # Used in _narrow_with_compose

        # FIXME Is this still needed?
        self.recipients = frozenset({self.user_id, view.model.user_id})

        super().__init__(controller,
                         caption=user['full_name'],
                         show_function=self._narrow_with_compose,
                         prefix_character=(color, state_marker),
                         text_color=color,
                         width=width,
                         count=count)
        if is_current_user:
            self.update_widget(('current_user', '(you)'), color)

    def _narrow_with_compose(self, button: Any) -> None:
        # Switches directly to composing with user
        # FIXME should we just narrow?
        self.controller.narrow_to_user(self)
        self._view.body.focus.original_widget.set_focus('footer')
        self._view.write_box.private_box_view(
            email=self.email,
            recipient_user_ids=[self.user_id]
        )


class TopicButton(TopButton):
    def __init__(self, stream_id: int, topic: str,
                 controller: Any, width: int=0, count: int=0) -> None:
        self.stream_name = controller.model.stream_dict[stream_id]['name']
        self.topic_name = topic
        self.stream_id = stream_id
        self.model = controller.model

        super().__init__(controller=controller,
                         caption=self.topic_name,
                         prefix_character='',
                         show_function=controller.narrow_to_topic,
                         width=width,
                         count=count)

        if controller.model.is_muted_topic(self.stream_id, self.topic_name):
            self.mark_muted()

    def mark_muted(self) -> None:
        self.update_widget(('muted',  MUTE_MARKER), 'muted')
    # TODO: Handle event-based approach for topic-muting.


class UnreadPMButton(urwid.Button):
    def __init__(self, user_id: int, email: str) -> None:
        self.user_id = user_id
        self.email = email


DecodedStream = TypedDict('DecodedStream', {
    'stream_id': Optional[int],
    'stream_name': Optional[str],
})

ParsedNarrowLink = TypedDict('ParsedNarrowLink', {
    'narrow': str,
    'stream': DecodedStream,
    'topic_name': str,
    'message_id': Optional[int],
}, total=False)


class MessageLinkButton(urwid.Button):
    def __init__(self, controller: Any, caption: str, link: str,
                 display_attr: Optional[str]) -> None:
        self.controller = controller
        self.model = self.controller.model
        self.view = self.controller.view
        self.link = link

        super().__init__('')
        self.update_widget(caption, display_attr)
        urwid.connect_signal(self, 'click', callback=self.handle_link)

    def update_widget(self, caption: str,
                      display_attr: Optional[str]=None) -> None:
        """
        Overrides the existing button widget for custom styling.
        """
        # Set cursor position next to len(caption) to avoid the cursor.
        icon = urwid.SelectableIcon(caption, cursor_position=len(caption) + 1)
        self._w = urwid.AttrMap(icon, display_attr, focus_map='selected')

    def handle_link(self, *_: Any) -> None:
        """
        Classifies and handles link.
        """
        server_url = self.model.server_url
        if self.link.startswith(urljoin(server_url, '/#narrow/')):
            self.handle_narrow_link()

    @staticmethod
    def _decode_stream_data(encoded_stream_data: str) -> DecodedStream:
        """
        Returns a dict with optional stream ID and stream name.
        """
        # Modern links come patched with the stream ID and '-' as delimiters.
        if re.match('^[0-9]+-', encoded_stream_data):
            stream_id, *_ = encoded_stream_data.split('-')
            # Given how encode_stream() in zerver/lib/url_encoding.py
            # replaces ' ' with '-' in the stream name, skip extracting the
            # stream name to avoid any ambiguity.
            return DecodedStream(stream_id=int(stream_id), stream_name=None)
        else:
            # Deprecated links did not start with the stream ID.
            stream_name = hash_util_decode(encoded_stream_data)
            return DecodedStream(stream_id=None, stream_name=stream_name)

    @staticmethod
    def _decode_message_id(message_id: str) -> Optional[int]:
        """
        Returns either the compatible near message ID or None.
        """
        try:
            return int(message_id)
        except ValueError:
            return None

    @classmethod
    def _parse_narrow_link(cls, link: str) -> ParsedNarrowLink:
        """
        Returns either a dict with narrow parameters for supported links or an
        empty dict.
        """
        # NOTE: The optional stream_id link version is deprecated. The extended
        # support is for old messages.
        # We expect the fragment to be one of the following types:
        # a. narrow/stream/[{stream_id}-]{stream-name}
        # b. narrow/stream/[{stream_id}-]{stream-name}/near/{message_id}
        # c. narrow/stream/[{stream_id}-]{stream-name}/topic/
        #    {encoded.20topic.20name}
        # d. narrow/stream/[{stream_id}-]{stream-name}/topic/
        #    {encoded.20topic.20name}/near/{message_id}
        fragments = urlparse(link.rstrip('/')).fragment.split('/')
        len_fragments = len(fragments)
        parsed_link = ParsedNarrowLink()

        if len_fragments == 3 and fragments[1] == 'stream':
            stream_data = cls._decode_stream_data(fragments[2])
            parsed_link = dict(narrow='stream', stream=stream_data)

        elif (len_fragments == 5 and fragments[1] == 'stream'
                and fragments[3] == 'topic'):
            stream_data = cls._decode_stream_data(fragments[2])
            topic_name = hash_util_decode(fragments[4])
            parsed_link = dict(narrow='stream:topic', stream=stream_data,
                               topic_name=topic_name)

        elif (len_fragments == 5 and fragments[1] == 'stream'
                and fragments[3] == 'near'):
            stream_data = cls._decode_stream_data(fragments[2])
            message_id = cls._decode_message_id(fragments[4])
            parsed_link = dict(narrow='stream:near', stream=stream_data,
                               message_id=message_id)

        elif (len_fragments == 7 and fragments[1] == 'stream'
                and fragments[3] == 'topic' and fragments[5] == 'near'):
            stream_data = cls._decode_stream_data(fragments[2])
            topic_name = hash_util_decode(fragments[4])
            message_id = cls._decode_message_id(fragments[6])
            parsed_link = dict(narrow='stream:topic:near', stream=stream_data,
                               topic_name=topic_name, message_id=message_id)

        return parsed_link

    def _validate_and_patch_stream_data(self,
                                        parsed_link: ParsedNarrowLink) -> str:
        """
        Validates stream data and patches the optional value in the nested
        DecodedStream dict.
        """
        stream_id = parsed_link['stream']['stream_id']
        stream_name = parsed_link['stream']['stream_name']
        assert (
            (stream_id is None and stream_name is not None)
            or (stream_id is not None and stream_name is None)
        )

        model = self.model
        # Validate stream ID and name.
        if ((stream_id and not model.is_user_subscribed_to_stream(stream_id))
                or (stream_name and not model.is_valid_stream(stream_name))):
            # TODO: Narrow to the concerned stream in a 'preview' mode or
            # report whether the stream id is invalid instead.
            return 'The stream seems to be either unknown or unsubscribed'

        # Patch the optional value.
        if not stream_id:
            stream_id = cast(int, model.stream_id_from_name(stream_name))
            parsed_link['stream']['stream_id'] = stream_id
        else:
            stream_name = cast(str, model.stream_dict[stream_id]['name'])
            parsed_link['stream']['stream_name'] = stream_name

        return ''

    def _validate_narrow_link(self, parsed_link: ParsedNarrowLink) -> str:
        """
        Returns either an empty string for a successful validation or an
        appropriate validation error.
        """
        if not parsed_link:
            return 'The narrow link seems to be either broken or unsupported'

        # Validate stream data.
        if 'stream' in parsed_link:
            error = self._validate_and_patch_stream_data(parsed_link)
            if error:
                return error

        # Validate topic name.
        if 'topic_name' in parsed_link:
            topic_name = parsed_link['topic_name']
            stream_id = parsed_link['stream']['stream_id']

            if topic_name not in self.model.topics_in_stream(stream_id):
                return 'Invalid topic name'

        # Validate message ID for near.
        if 'near' in parsed_link['narrow']:
            message_id = parsed_link.get('message_id')

            if message_id is None:
                return 'Invalid message ID'

        return ''

    def _switch_narrow_to(self, parsed_link: ParsedNarrowLink) -> None:
        """
        Switches narrow via narrow_to_* methods.
        """
        narrow = parsed_link['narrow']
        if 'stream' == narrow:
            self.stream_id = parsed_link['stream']['stream_id']
            self.stream_name = parsed_link['stream']['stream_name']
            self.controller.narrow_to_stream(self)
        elif 'stream:near' == narrow:
            self.stream_id = parsed_link['stream']['stream_id']
            self.stream_name = parsed_link['stream']['stream_name']
            self.message = dict(id=parsed_link['message_id'])
            self.controller.narrow_to_stream(self)
        elif 'stream:topic' == narrow:
            self.stream_id = parsed_link['stream']['stream_id']
            self.stream_name = parsed_link['stream']['stream_name']
            self.topic_name = parsed_link['topic_name']
            self.controller.narrow_to_topic(self)
        elif 'stream:topic:near' == narrow:
            self.stream_id = parsed_link['stream']['stream_id']
            self.stream_name = parsed_link['stream']['stream_name']
            self.topic_name = parsed_link['topic_name']
            self.message = dict(id=parsed_link['message_id'])
            self.controller.narrow_to_topic(self)

    def handle_narrow_link(self) -> None:
        """
        Narrows to the respective narrow if the narrow link is valid or updates
        the footer with an appropriate validation error message.
        """
        parsed_link = self._parse_narrow_link(self.link)
        error = self._validate_narrow_link(parsed_link)

        if error:
            self.view.set_footer_text(' {}'.format(error), duration=3)
        else:
            self._switch_narrow_to(parsed_link)

            # Exit pop-up if MessageLinkButton exists in one.
            if isinstance(self.controller.loop.widget, urwid.Overlay):
                self.controller.exit_popup()


class EditModeButton(urwid.Button):
    def __init__(self, controller: Any, width: int) -> None:
        self.controller = controller
        self.width = width
        super().__init__(label="",
                         on_press=controller.show_topic_edit_mode)
        self.set_selected_mode('change_later')  # set default mode

    def set_selected_mode(self, mode: str) -> None:
        self.mode = mode
        self._w = urwid.AttrMap(urwid.SelectableIcon(
            edit_mode_captions[self.mode], self.width), None, 'selected')



import argparse
import configparser
import logging
import os
import stat
import sys
import traceback
from os import path, remove
from typing import Any, Dict, List, Optional, Tuple

import requests
from urwid import display_common, set_encoding

from zulipterminal.config.themes import (
    THEMES,
    aliased_themes,
    all_themes,
    complete_and_incomplete_themes,
    theme_with_monochrome_added,
)
from zulipterminal.core import Controller
from zulipterminal.model import ServerConnectionFailure
from zulipterminal.version import ZT_VERSION


TRACEBACK_LOG_FILENAME = 'zulip-terminal-tracebacks.log'
API_CALL_LOG_FILENAME = 'zulip-terminal-API-requests.log'

# Create a logger for this application
zt_logger = logging.getLogger(__name__)
zt_logger.setLevel(logging.DEBUG)
zt_logfile_handler = logging.FileHandler(
    TRACEBACK_LOG_FILENAME,
    delay=True,  # Don't open the file until there's a logging event
)
zt_logger.addHandler(zt_logfile_handler)

# Route requests details (API calls) to separate file
requests_logger = logging.getLogger("urllib3")
requests_logger.setLevel(logging.DEBUG)

# These should be the defaults without config file or command-line overrides
DEFAULT_SETTINGS = {
    'theme': 'zt_dark',
    'autohide': 'no_autohide',
    'notify': 'disabled',
    'footlinks': 'enabled',
    'color-depth': '256',
}


def in_color(color: str, text: str) -> str:
    color_for_str = {
        'red': '1',
        'green': '2',
        'yellow': '3',
        'blue': '4',
        'purple': '5',
        'cyan': '6',
    }
    # We can use 3 instead of 9 if high-contrast is eg. less compatible?
    return "\033[9{}m{}\033[0m".format(color_for_str[color], text)


def exit_with_error(error_message: str, *,
                    helper_text: str="", error_code: int=1) -> None:
    print(in_color("red", error_message))
    if helper_text:
        print(helper_text)
    sys.exit(error_code)


def parse_args(argv: List[str]) -> argparse.Namespace:
    description = """
        Starts Zulip-Terminal.
        """
    formatter_class = argparse.RawDescriptionHelpFormatter
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=formatter_class)

    parser.add_argument('--config-file', '-c',
                        action='store',
                        help='config file downloaded from your zulip '
                             'organization.(e.g. ~/zuliprc)')
    parser.add_argument('--theme', '-t',
                        help='choose color theme (default {}).'
                             .format(DEFAULT_SETTINGS['theme']))
    parser.add_argument('--list-themes',
                        action="store_true",
                        help='list all the color themes.')
    parser.add_argument('--color-depth',
                        choices=['1', '16', '256'],
                        help="Force the color depth (default {})."
                             .format(DEFAULT_SETTINGS['color-depth']))
    # debug mode
    parser.add_argument("-d",
                        "--debug",
                        action="store_true",
                        help="Start zulip terminal in debug mode.")

    parser.add_argument('--profile', dest='profile',
                        action="store_true",
                        default=False, help='Profile runtime.')

    autohide_group = parser.add_mutually_exclusive_group()
    autohide_group.add_argument('--autohide', dest='autohide', default=None,
                                action="store_const", const='autohide',
                                help='Autohide list of users and streams.')

    autohide_group.add_argument('--no-autohide', dest='autohide', default=None,
                                action="store_const", const='no_autohide',
                                help='Don\'t autohide list of '
                                     'users and streams.')

    parser.add_argument('-e', '--explore', action='store_true',
                        help='Do not mark messages as read in the session.')

    parser.add_argument('-v',
                        '--version',
                        action='store_true',
                        default=False,
                        help='Print zulip-terminal version and exit')

    return parser.parse_args(argv)


def styled_input(label: str) -> str:
    return input(in_color('blue', label))


def get_login_id(realm_url: str) -> str:
    res_json = requests.get(url=realm_url + '/api/v1/server_settings').json()
    require_email_format_usernames = res_json['require_email_format_usernames']
    email_auth_enabled = res_json['email_auth_enabled']

    if not require_email_format_usernames and email_auth_enabled:
        label = 'Email or Username: '
    elif not require_email_format_usernames:
        label = 'Username: '
    else:
        # TODO: Validate Email address
        label = 'Email: '

    return styled_input(label)


def get_api_key(realm_url: str) -> Tuple[requests.Response, str]:
    from getpass import getpass

    login_id = get_login_id(realm_url)
    password = getpass(in_color('blue', "Password: "))
    response = requests.post(
        url=realm_url + '/api/v1/fetch_api_key',
        data={
            'username': login_id,
            'password': password,
        }
    )
    return response, login_id


def fetch_zuliprc(zuliprc_path: str) -> None:
    print(in_color('red', "zuliprc file was not found at " + zuliprc_path)
          + "\nPlease enter your credentials to login into your"
            " Zulip organization."
            "\n"
            "\nNOTE: The " + in_color('blue', "Zulip URL")
          + " is where you would go in a web browser to log in to Zulip."
            "\nIt often looks like one of the following:"
            "\n   " + in_color('green', "your-org.zulipchat.com")
          + " (Zulip cloud)"
            "\n   " + in_color('green', "zulip.your-org.com")
          + " (self-hosted servers)"
            "\n   " + in_color('green', "chat.zulip.org")
          + " (the Zulip community server)")
    realm_url = styled_input('Zulip URL: ')
    if realm_url.startswith("localhost"):
        realm_url = "http://" + realm_url
    elif not realm_url.startswith("http"):
        realm_url = "https://" + realm_url
    # Remove trailing "/"s from realm_url to simplify the below logic
    # for adding "/api"
    realm_url = realm_url.rstrip("/")
    res, login_id = get_api_key(realm_url)

    while res.status_code != 200:
        print(in_color('red', "\nIncorrect Email(or Username) or Password!\n"))
        res, login_id = get_api_key(realm_url)

    save_zuliprc_failure = _write_zuliprc(
        zuliprc_path,
        login_id=login_id,
        api_key=str(res.json()['api_key']),
        server_url=realm_url,
    )
    if not save_zuliprc_failure:
        print('Generated API key saved at ' + zuliprc_path)
    else:
        exit_with_error(save_zuliprc_failure)


def _write_zuliprc(to_path: str, *,
                   login_id: str, api_key: str, server_url: str) -> str:
    """
    Writes a zuliprc file, returning a non-empty error string on failure
    Only creates new private files; errors if file already exists
    """
    try:
        with open(os.open(to_path,
                          os.O_CREAT | os.O_WRONLY | os.O_EXCL,
                          0o600),
                  'w') as f:
            f.write('[api]\nemail={}\nkey={}\nsite={}'
                    .format(login_id, api_key, server_url))
        return ""
    except FileExistsError as ex:
        return "zuliprc already exists at {}".format(to_path)
    except OSError as ex:
        return ("{}: zuliprc could not be created at {}"
                .format(ex.__class__.__name__, to_path))


def parse_zuliprc(zuliprc_str: str) -> Dict[str, Any]:
    zuliprc_path = path.expanduser(zuliprc_str)
    while not path.exists(zuliprc_path):
        try:
            fetch_zuliprc(zuliprc_path)
        # Invalid user inputs (e.g. pressing arrow keys) may cause ValueError
        except (OSError, ValueError):
            # Remove zuliprc file if created.
            if path.exists(zuliprc_path):
                remove(zuliprc_path)
            print(in_color('red',
                           "\nInvalid Credentials, Please try again!\n"))
        except EOFError:
            # Assume that the user pressed Ctrl+D and continue the loop
            print("\n")

    mode = os.stat(zuliprc_path).st_mode
    is_readable_by_group_or_others = mode & (stat.S_IRWXG | stat.S_IRWXO)

    if is_readable_by_group_or_others:
        print(in_color(
            'red',
            "ERROR: Please ensure your zuliprc is NOT publicly accessible:\n"
            "  {0}\n"
            "(it currently has permissions '{1}')\n"
            "This can often be achieved with a command such as:\n"
            "  chmod og-rwx {0}\n"
            "Consider regenerating the [api] part of your zuliprc to ensure "
            "your account is secure.".format(zuliprc_path, stat.filemode(mode))
        ))
        sys.exit(1)

    zuliprc = configparser.ConfigParser()

    try:
        res = zuliprc.read(zuliprc_path)
        if len(res) == 0:
            exit_with_error("Could not access zuliprc file at " + zuliprc_path)
    except configparser.MissingSectionHeaderError:
        exit_with_error("Failed to parse zuliprc file at " + zuliprc_path)

    # Initialize with default settings
    NO_CONFIG = 'with no config'
    settings = {
        setting: (default, NO_CONFIG)
        for setting, default in DEFAULT_SETTINGS.items()
    }

    if 'zterm' in zuliprc:
        config = zuliprc['zterm']
        ZULIPRC_CONFIG = 'in zuliprc file'
        for conf in config:
            settings[conf] = (config[conf], ZULIPRC_CONFIG)

    return settings


def list_themes() -> str:
    available_themes = all_themes()
    text = "The following themes are available:\n"
    for theme in available_themes:
        suffix = ""
        if theme == DEFAULT_SETTINGS['theme']:
            suffix += "[default theme]"
        text += "  {} {}\n".format(theme, suffix)
    return text + (
        "Specify theme in zuliprc file or override "
        "using -t/--theme options on command line."
    )


def main(options: Optional[List[str]]=None) -> None:
    """
    Launch Zulip Terminal.
    """

    argv = options if options is not None else sys.argv[1:]
    args = parse_args(argv)

    set_encoding('utf-8')

    if args.debug:
        print("NOTE: Debug mode enabled; API calls being logged to {}."
              .format(in_color("blue", API_CALL_LOG_FILENAME)))
        requests_logfile_handler = logging.FileHandler(API_CALL_LOG_FILENAME)
        requests_logger.addHandler(requests_logfile_handler)
    else:
        requests_logger.addHandler(logging.NullHandler())

    if args.profile:
        import cProfile
        prof = cProfile.Profile()
        prof.enable()

    if args.version:
        print('Zulip Terminal ' + ZT_VERSION)
        sys.exit(0)

    if args.list_themes:
        print(list_themes())
        sys.exit(0)

    if args.config_file:
        zuliprc_path = args.config_file
    else:
        zuliprc_path = '~/zuliprc'

    try:
        zterm = parse_zuliprc(zuliprc_path)

        if args.autohide:
            zterm['autohide'] = (args.autohide, 'on command line')

        if args.theme:
            theme_to_use = (args.theme, 'on command line')
        else:
            theme_to_use = zterm['theme']

        available_themes = all_themes()
        theme_aliases = aliased_themes()
        is_valid_theme = (
            theme_to_use[0] in available_themes
            or theme_to_use[0] in theme_aliases
        )
        if not is_valid_theme:
            exit_with_error(
                "Invalid theme '{}' was specified {}.".format(*theme_to_use),
                helper_text=list_themes(),
            )
        if theme_to_use[0] not in available_themes:
            # theme must be an alias, as it is valid
            real_theme_name = theme_aliases[theme_to_use[0]]
            theme_to_use = (
                real_theme_name,
                "{} (by alias '{}')".format(theme_to_use[1], theme_to_use[0])
            )

        if args.color_depth:
            zterm['color-depth'] = (args.color_depth, 'on command line')

        color_depth = int(zterm['color-depth'][0])

        print("Loading with:")
        print("   theme '{}' specified {}.".format(*theme_to_use))
        complete, incomplete = complete_and_incomplete_themes()
        if theme_to_use[0] in incomplete:
            print(in_color('yellow',
                           "   WARNING: Incomplete theme; "
                           "results may vary!\n"
                           "      (you could try: {})".
                           format(", ".join(complete))))
        print("   autohide setting '{}' specified {}."
              .format(*zterm['autohide']))
        print("   footlinks setting '{}' specified {}."
              .format(*zterm['footlinks']))
        print("   color depth setting '{}' specified {}."
              .format(*zterm['color-depth']))
        # For binary settings
        # Specify setting in order True, False
        valid_settings = {
            'autohide': ['autohide', 'no_autohide'],
            'notify': ['enabled', 'disabled'],
            'footlinks': ['enabled', 'disabled'],
            'color-depth': ['1', '16', '256']
        }
        boolean_settings = dict()  # type: Dict[str, bool]
        for setting, valid_values in valid_settings.items():
            if zterm[setting][0] not in valid_values:
                helper_text = ["Valid values are:"] + [
                    "  {}".format(option) for option in valid_values
                ] + ["Specify the {} option in zuliprc file.".format(setting)]
                exit_with_error(
                    ("Invalid {} setting '{}' was specified {}."
                     .format(setting, *zterm[setting])),
                    helper_text="\n".join(helper_text),
                )
            if setting == 'color-depth':
                break
            boolean_settings[setting] = (zterm[setting][0] == valid_values[0])

        if color_depth == 1:
            theme_data = theme_with_monochrome_added(THEMES[theme_to_use[0]])
        else:
            theme_data = THEMES[theme_to_use[0]]

        Controller(zuliprc_path,
                   theme_to_use[0],
                   theme_data,
                   color_depth,
                   args.explore,
                   **boolean_settings).main()
    except ServerConnectionFailure as e:
        # Acts as separator between logs
        zt_logger.info("\n\n" + str(e) + "\n\n")
        zt_logger.exception(e)
        exit_with_error("\nError connecting to Zulip server: {}.".format(e))
    except (display_common.AttrSpecError, display_common.ScreenError) as e:
        # NOTE: Strictly this is not necessarily just a theme error
        # FIXME: Add test for this - once loading takes place after UI setup

        # Acts as separator between logs
        zt_logger.info("\n\n" + str(e) + "\n\n")
        zt_logger.exception(e)
        exit_with_error("\nPossible theme error: {}.".format(e))
    except Exception as e:
        zt_logger.info("\n\n" + str(e) + "\n\n")
        zt_logger.exception(e)
        if args.debug:
            sys.stdout.flush()
            traceback.print_exc(file=sys.stderr)
            run_debugger = input("Run Debugger? (y/n): ")
            if run_debugger in ["y", "Y", "yes"]:
                # Open PUDB Debugger
                import pudb
                pudb.post_mortem()

        if hasattr(e, 'extra_info'):
            print("\n" + in_color("red", e.extra_info),    # type: ignore
                  file=sys.stderr)

        print(in_color("red", "\nZulip Terminal has crashed!"
                       "\nPlease refer to " + TRACEBACK_LOG_FILENAME
                       + " for full log of the error."), file=sys.stderr)
        print("You can ask for help at:", file=sys.stderr)
        print("https://chat.zulip.org/#narrow/stream/206-zulip-terminal",
              file=sys.stderr)
        print("\nThanks for using the Zulip-Terminal interface.\n")
        sys.stderr.flush()

    finally:
        if args.profile:
            prof.disable()
            import tempfile
            with tempfile.NamedTemporaryFile(prefix="zulip_term_profile.",
                                             suffix=".dat",
                                             delete=False) as profile_file:
                profile_path = profile_file.name
            # Dump stats only after temporary file is closed (for Win NT+ case)
            prof.dump_stats(profile_path)
            print(
                "Profile data saved to {0}.\n"
                "You can visualize it using e.g. `snakeviz {0}`"
                .format(profile_path)
            )

        sys.exit(1)


if __name__ == '__main__':
    main()
